/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */
#include <filesystem>
#include <vector>
#include <map>
#include <string>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <stdexcept>
#include <matplot/matplot.h>
#include "runtimeInfo/runtimeInfo.h"
#include "moduleBasics/html.h"
#include "moduleBasics/svg.h"
#include "tawCalculatePolar.h"

namespace fs = std::filesystem;
namespace helpers {

// Define colors
constexpr svg::Colors::Color_t Grey{128, 128, 128};
constexpr svg::Colors::Color_t BlueLight{100, 160, 200};
constexpr svg::Colors::Color_t BlueDark{0, 51, 89};
constexpr svg::Colors::Color_t Orange{227, 114, 43};
constexpr svg::Colors::Color_t Green{162, 173, 0};

// Normalize color values for Matplot++
inline auto normalize_color(const svg::Colors::Color_t& color) -> std::array<float, 3> {
    return {std::get<0>(color) / 255.0F, std::get<1>(color) / 255.0F, std::get<2>(color) / 255.0F};
}

// Configure axes properties
inline void configure_axes(matplot::axes_handle& ax, const std::string& title,
                            const std::string& xlabel, const std::string& ylabel) {
    ax->title(title);
    ax->font_size(16);
    ax->xlabel(xlabel);
    ax->x_axis().label_font_size(16);
    ax->ylabel(ylabel);
    ax->y_axis().label_font_size(16);
    ax->grid(true);
}

// Configure legend properties
inline void configure_legend(matplot::axes_handle& ax) {
    if (auto legend = ax->legend()) {
        legend->location(matplot::legend::general_alignment::bottomright);
        legend->inside(false);
        legend->box(true);
        legend->font_size(16);
    }
}

// Plot datasets
inline void plot_datasets(matplot::axes_handle& ax,
                           const std::map<std::string, std::pair<std::vector<double>, std::vector<double>>>& data) {
    const std::array<svg::Colors::Color_t, 5> plot_colors = {BlueLight, BlueDark, Orange, Green, Grey};
    size_t color_index = 0;

    for (const auto& [label, coefficients] : data) {
        const auto& x_data = coefficients.first;
        const auto& y_data = coefficients.second;

        if (!x_data.empty() && !y_data.empty()) {
            auto line_color = normalize_color(plot_colors[color_index % plot_colors.size()]);
            ax->plot(x_data, y_data)
                ->line_width(2)
                .color(line_color)
                .display_name(label);
            // Dynamically create or update the legend
            if (ax->legend() == nullptr) {
                ax->legend({label});
            } else {
                ax->legend()->strings().push_back(label);
            }
            ++color_index;
        }
    }
}

// Prepare the plot filename and path
inline auto prepare_plot_path(const std::string& plot_dir, const std::string& plot_filename) -> std::string {
    auto plot_path = fs::path(plot_dir) / plot_filename;
    std::string plot_path_string = plot_path.string();
    std::replace(plot_path_string.begin(), plot_path_string.end(), '\\', '/');
    return plot_path_string;
}

// Trim whitespace from a string
inline std::string trim_whitespace(const std::string& str) {
    auto start = str.find_first_not_of(" \t\r\n");
    auto end = str.find_last_not_of(" \t\r\n");
    return (start == std::string::npos || end == std::string::npos) ? "" : str.substr(start, end - start + 1);
}

}  // namespace helpers

// Parse a CSV file
auto parse_csv_file(const std::string& file_path, const std::string& column_x,
                    const std::string& column_y, char delimiter = ';')
    -> std::pair<std::vector<double>, std::vector<double>> {
    std::ifstream csv_file(file_path);
    if (!csv_file.is_open()) {
        myRuntimeInfo->warn << "Could not open CSV file: " + file_path << std::endl;
    }

    std::string line;
    std::vector<std::string> headers;
    std::vector<double> x_data, y_data;
    bool is_header = true;

    while (std::getline(csv_file, line)) {
        if (line.empty() || line[0] == '#') continue;  // Skip comments and empty lines

        std::istringstream ss(line);
        std::string token;

        if (is_header) {
            while (std::getline(ss, token, delimiter)) {
                headers.push_back(helpers::trim_whitespace(token));
            }
            is_header = false;
            continue;
        }

        std::map<std::string, double> row_data;
        size_t column_index = 0;

        while (std::getline(ss, token, delimiter)) {
            try {
                double value = std::stod(helpers::trim_whitespace(token));
                row_data[headers[column_index]] = value;
            } catch (...) {
                row_data[headers[column_index]] = 0.0;  // Default invalid data to 0
            }
            ++column_index;
        }

        try {
            x_data.push_back(row_data.at(column_x));
            y_data.push_back(row_data.at(column_y));
        } catch (const std::out_of_range&) {
            myRuntimeInfo->warn << "Specified columns not found in CSV: " + column_x + ", " + column_y << std::endl;
        }
    }

    return {x_data, y_data};
}

fs::path generate_plot(
    const std::map<std::string, std::pair<std::vector<double>, std::vector<double>>>& data,
    const std::string& plot_dir, const std::string& html_dir,
    const std::string& filename, const std::string& title,
    const std::string& xlabel, const std::string& ylabel) {

    if (data.empty()) {
        myRuntimeInfo->warn << "No data provided for plotting." << std::endl;
    }

    auto plot_path_string = helpers::prepare_plot_path(plot_dir, filename);

    auto f = matplot::figure(true);
    f->size(800, 600);
    auto ax = f->current_axes();
    ax->hold(true);
    if (title == "Polar Curve (clean configuration)") {
        ax->xlim({0, 0.08});
    }
    helpers::plot_datasets(ax, data);
    helpers::configure_axes(ax, title, xlabel, ylabel);
    helpers::configure_legend(ax);

    matplot::save(f, plot_path_string, "svg");
    return fs::relative(fs::path(plot_path_string), fs::path(html_dir));
}

void tawCalculatePolar::set_html_body() {
    const auto& data = myDefaultStrategy->get_default_data();
    const auto& config = myDefaultStrategy->get_taw_config();

    // Begin HTML Body
    report_.htmlReportStream() << "<div class=\"box data\">\n";

    // Headline for Data Section
    report_.htmlReportStream() << "<h2>Data</h2>\n";

    // Reference Values Table
    report_.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Reference values</caption>\n"
        << "<thead>\n<tr>\n"
        << "<th>Parameter</th><th>Symbol</th><th>Unit</th><th>Value</th>\n"
        << "</tr>\n</thead>\n<tbody>\n"
        << std::format("<tr><td>Wing half span</td>"
                       "<td>b</td><td>m</td><td>{:.2f}</td></tr>\n", data.span.value())
        << std::format("<tr><td>Mean aerodynamic chord</td>"
                       "<td>MAC</td><td>m</td><td>{:.2f}</td></tr>\n", data.MAC.value())
        << std::format("<tr><td>Reference area</td>"
                       "<td>S<sub>ref</sub></td><td>m2</td><td>{:.1f}</td></tr>\n", data.S_ref.value())
        << std::format("<tr><td>Neutral point</td>"
                       "<td>x</td><td>m</td><td>{:.2f}</td></tr>\n", data.neutral_point_xCoordinate.value())
        << "</tbody>\n</table>\n";

    // Design Cruise Conditions Table
    report_.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Design cruise conditions</caption>\n"
        << "<thead>\n<tr>\n"
        << "<th>Parameter</th><th>Symbol</th><th>Unit</th><th>Value</th>\n"
        << "</tr>\n</thead>\n<tbody>\n"
        << std::format("<tr><td>Mach number</td>"
                       "<td>Ma<sub>Cr</sub></td><td>-</td><td>{:.2f}</td></tr>\n", data.initialCruiseMachNumber)
        << std::format("<tr><td>Best L/D</td>"
                       "<td>L/D<sub>,opt,cruise</sub></td><td>-</td><td>{:.2f}</td></tr>\n",
                       data.L_over_D_optimumCruise.value())
        << std::format("<tr><td>Optimum lift coefficient at cruise (for L/D max)</td>"
                       "<td>C<sub>L,opt,cruise</sub></td><td>-</td><td>{:.3f}</td></tr>\n",
                       data.C_L_optimumCruise.value())
        << std::format("<tr><td>Total drag at optimum lift coefficient</td>"
                       "<td>C<sub>D,total,opt,cruise</sub></td><td>-</td><td>{:.5f}</td></tr>\n",
                       data.C_Dtotal_optimumCruise.value())
        << std::format("<tr><td>Induced drag at optimum lift coefficient</td>"
                       "<td>C<sub>D,induced,opt,cruise</sub></td><td>-</td><td>{:.5f}</td><td>({:.2f} \%)</td></tr>\n",
                       data.C_Dind_optimumCruise.value(),
                       Rounding(data.C_Dind_optimumCruise.value() / data.C_Dtotal_optimumCruise.value() * 100., 2.))
        << std::format("<tr><td>Viscous drag at optimum lift coefficient</td>"
                       "<td>C<sub>D,viscous,opt,cruise</sub></td><td>-</td><td>{:.5f}</td><td>({:.2f} \%)</td></tr>\n",
                       data.C_Dvisc_optimumCruise.value(),
                       Rounding(data.C_Dvisc_optimumCruise.value() / data.C_Dtotal_optimumCruise.value() * 100., 2.))
        << std::format("<tr><td>Wave drag at optimum lift coefficient</td>"
                       "<td>C<sub>D,wave,opt,cruise</sub></td><td>-</td><td>{:.5f}</td><td>({:.2f} \%)</td></tr>\n",
                       data.C_Dwave_optimumCruise.value(),
                       Rounding(data.C_Dwave_optimumCruise.value() / data.C_Dtotal_optimumCruise.value() * 100., 2.))
        << std::format("<tr><td>Moment coefficient at optimum lift coefficient</td>"
                       "<td>C<sub>M,opt,cruise</sub></td><td>-</td><td>{:.5f}</td></tr>\n",
                       data.C_M_optimumCruise.value())
        << "</tbody>\n</table>\n";

    // Lift Coefficients Table
    report_.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Lift coefficients</caption>\n"
        << "<thead>\n<tr>\n"
        << "<th>Parameter</th><th>Symbol</th><th>Unit</th><th>Value</th>\n"
        << "</tr>\n</thead>\n<tbody>\n"
        << std::format("<tr><td>Max. lift coeff. landing</td>"
                       "<td>C<sub>L,max,LD</sub></td><td>-</td><td>{:.3f}</td></tr>\n", data.CL_max_landing.value())
        << std::format("<tr><td>Max. lift coeff. take-off</td>"
                       "<td>C<sub>L,max,TO</sub></td><td>-</td><td>{:.3f}</td></tr>\n", data.C_LmaxTO.value())
        << "</tbody>\n</table>\n";
    // Closing Table
    report_.htmlReportStream() << "</tbody>\n</table>\n";

    // Closing Data Section
    report_.htmlReportStream() << "</div>\n";
    if (rtIO_->plotOn) {
        /* open box plot div */
        report_.htmlReportStream() << "<div class=\"box plot\">\n";
        /* Add headline Data*/
        report_.htmlReportStream() << "<h2>Plots</h2>\n";;

        // Flight Condition
        std::vector<std::pair<std::string, std::string>> polar_files;
        for (size_t i = 0; i < config.flightConditions.size(); i++) {
            std::string mach_number = std::format("{:.2f}",config.flightConditions[i].Mach);
            std::string converted = mach_number;
            mach_number = "Mach " + mach_number;

            converted.erase(std::remove(converted.begin(), converted.end(), '.'), converted.end());
            std::string filename = std::string("tawCalculatePolar_M") + converted + "_design_000_plot.csv";
            polar_files.push_back({filename, mach_number});
        }

        std::vector<std::pair<std::string, std::string>> drag_components = {
            {"C_Dwave", "C_{D,wave}"},
            {"C_Dind", "C_{D,ind}"},
            {"C_Dvisc", "C_{D,visc}"},
            {"C_Dtotal", "C_{D,total}"}
        };

        // Cruise Speed
        double machcruise = data.initialCruiseMachNumber;
        std::string mach_number_string = std::to_string(machcruise);
        mach_number_string.erase(mach_number_string.length() - 4);
        mach_number_string.erase(std::remove(mach_number_string.begin(), mach_number_string.end(), '.'), mach_number_string.end());

        std::map<std::string, std::pair<std::vector<double>, std::vector<double>>> polar_data;
        std::map<std::string, std::pair<std::vector<double>, std::vector<double>>> lift_curve_data;
        std::map<std::string, std::pair<std::vector<double>, std::vector<double>>> lift_to_drag_data;
        std::map<std::string, std::pair<std::vector<double>, std::vector<double>>> wing_cruise_drag_data;

        try {
            for (const auto& [filename, label] : polar_files) {
                auto file_path = fs::path(rtIO_->getCsvFilesDir()) / filename;
                polar_data[label] = parse_csv_file(file_path.string(), "C_Dtotal", "C_L");
                lift_curve_data[label] = parse_csv_file(file_path.string(), "AoA", "C_L");
                lift_to_drag_data[label] = parse_csv_file(file_path.string(), "C_L", "L/D");
            }

            // Drag-Plot data
            auto file_path = fs::path(rtIO_->getCsvFilesDir()) / ("tawCalculatePolar_M"+ mach_number_string +"_design_000_plot.csv");

            for (const auto& [component_column, component] : drag_components) {
                wing_cruise_drag_data[component] = parse_csv_file(file_path.string(), component_column, "C_L");
            }

            // Polar-clean
            auto polar_clean_plot_path = generate_plot(
                polar_data,
                rtIO_->getPlotDir(),
                rtIO_->getHtmlDir(),
                "polar_clean.svg",
                "Polar Curve (clean configuration)",
                "C_D",
                "C_L");

            // Lift-Curve polar
            auto lift_curve_plot_path = generate_plot(
                lift_curve_data,
                rtIO_->getPlotDir(),
                rtIO_->getHtmlDir(),
                "lift_curve_clean.svg",
                "Lift curve (clean configuration)",
                "AoA",
                "C_L");

            // Lift-to-drag polar

            auto lift_to_drag_ratio_plot_path = generate_plot(
                lift_to_drag_data,
                rtIO_->getPlotDir(),
                rtIO_->getHtmlDir(),
                "lift_to_drag_clean.svg",
                "Lift-to-drag-ratio (clean configuration)",
                "C_L",
                "L/D");

            // Wing cruise drag polar
            auto wing_cruise_drag_path = generate_plot(
                wing_cruise_drag_data,
                rtIO_->getPlotDir(),
                rtIO_->getHtmlDir(),
                "wing_cruise_drag_clean.svg",
                "Wing cruise drag polar (clean configuration)",
                "C_D",
                "C_L");

            report_.htmlReportStream() << std::format("<img src=\"{}\" alt=\"Lift vs. Drag\"/><br>\n",
            polar_clean_plot_path.string());
            report_.htmlReportStream() << std::format("<img src=\"{}\" alt=\"Lift curve\"/><br>\n",
            lift_curve_plot_path.string());
            report_.htmlReportStream() << std::format("<img src=\"{}\" alt=\"Lift-to-drag-ratio\"/><br>\n",
            lift_to_drag_ratio_plot_path.string());
            report_.htmlReportStream() << std::format("<img src=\"{}\" alt=\"Drag-components\"/><br>\n",
            wing_cruise_drag_path.string());
        } catch (const std::exception& e) {
            myRuntimeInfo->warn << "Error generating plots: " << e.what() << std::endl;
        }

        report_.htmlReportStream() << "</div>\n";
    }
}
